"""Biobank Dataset class."""

import glob
import itertools
import re
import shutil

import dask.dataframe as dd
import numpy as np
import pandas as pd
import pyarrow.parquet as pq
from dask.diagnostics import ProgressBar

from biobank.config import settings
from biobank.dictionary import Dictionary
from biobank.importers import ImportManager


class Dataset:
    """Biobank Dataset class."""

    filename = "dataset.parquet"

    def __init__(self):
        """Constructor."""
        path = settings.path.absolute()
        self.path = path / self.filename
        self.dictionary = Dictionary(path)

    def delete(self) -> None:
        """Delete the dataset.

        Returns:
            None
        """
        if self.path.is_file():
            self.path.unlink()
        else:
            shutil.rmtree(str(self.path))

    def load(self, **kwargs) -> dd.DataFrame:
        """Loads a previously imported Biobank dataset.

        Args:
            **kwargs: Dictionary of keyword arguments

        Returns:
            A Dask DataFrame object
        """
        return dd.read_parquet(str(self.path), **kwargs)

    def save(self, data, schema, **kwargs) -> None:
        """Saves a Biobank dataset.

        Args:
            data: Biobank dataset as a Dask Dataframe
            schema: Parquet schema
            **kwargs: Dictionary of keyword arguments

        Returns:
            None
        """
        print(f"saving dataset to {self.path}")
        data.to_parquet(
            self.path, schema=schema, compression="snappy", engine="pyarrow"
        )

    def load_metadata(self) -> pq.FileMetaData:
        """Loads metadata for the Biobank dataset.

        Returns:
            A FileMetadata object
        """
        return pq.read_metadata(self.path / "_common_metadata")

    def filter_dictionary(self, fields, search):
        return self.dictionary.filter(self.match_fields(fields), search)

    def match_fields(self, fields=None):
        columns = self.load_metadata().schema.names
        if not fields:
            return columns

        pattern = "|".join(map(lambda field: rf"({field}-\d+\.\d+)", fields))
        pattern = rf"^{pattern}$"
        columns = filter(lambda field: re.match(pattern, field), columns)
        return list(columns)

    def import_dataset(self, path, dictionary) -> None:
        """Import a dataset.

        Args:
            path: URL or local path of dataset to import

        Returns:
            None
        """
        self.dictionary.load(dictionary, download=True)
        with ProgressBar():
            import_manager = ImportManager()
            data, schema = import_manager.import_dataset(self.dictionary, path)
        with ProgressBar():
            self.save(data, schema)

    def exclude(self, files):
        files = list(
            itertools.chain(*map(lambda path: glob.glob(str(path)), files))
        )
        exclude = (
            pd.concat(map(lambda file: pd.read_csv(file, header=None), files))
            .set_index(0)
            .index.unique()
        )

        print(f"removing {len(exclude)} records from {len(files)} files:")
        for file in files:
            print(f"- {file}")

        with ProgressBar():
            data = self.load()
            data = data.loc[~data.index.isin(exclude)]
            schema = self.load_metadata().schema.to_arrow_schema()
            schema = {
                name: dtype for name, dtype in zip(schema.names, schema.types)
            }
            self.save(data, schema)

    def select(self, fields=None, limit=None) -> pd.DataFrame:
        """Select specific fields from the Biobank dataset.

        Args:
            fields: List of fields to select
            limit: Number of rows to select

        Returns:
            A Pandas DataFrame
        """
        if fields:
            fields = self.match_fields(fields)
            if not len(fields):
                return pd.DataFrame()
        else:
            fields = None

        with ProgressBar():
            dataset = self.load(columns=fields, use_threads=True)
            if limit:
                dataset = dataset.loc[
                    dataset.index.isin(dataset.index.head(limit))
                ]

            dataset = dataset.replace(
                to_replace={
                    col: {np.nan: ""}
                    for col in dataset.select_dtypes(
                        [np.float64, np.datetime64, object]
                    ).columns
                }
            )
            dataset = dataset.compute()

        return dataset



"""This file provides a meta class for a singleton object."""

from typing import Any


class Singleton(type):
    """Singleton class.

    A singleton pattern that can be used as a metaclass to ensure only one
    instance of the target class is ever instantiated
    """

    def __init__(self, name, bases, dic):
        self._instance = None
        super().__init__(name, bases, dic)

    def __call__(cls, *args, **kwargs) -> Any:
        """Metcalsss __call__ method.

        Overrides the __call__ function so we can eiter return an existing
        instance or create a new one if necessary.

        Args:
            *args: List of positional arguments
            **kwargs: Dictionary of keywork arguments

        Returns:
            An instance of the singleton class
        """
        if not cls._instance:
            cls._instance = super().__call__(*args, **kwargs)
        return cls._instance



"""Dictionary class for managing data dictionary."""

from pathlib import Path
from typing import Any, List

import dask.dataframe as dd
import pandas as pd
import pyarrow as pa
from dask.diagnostics import ProgressBar

from biobank.config import settings


class Dictionary:
    """Dictionary class for managing data dictionary."""

    field_types = {
        "Integer": pa.int64(),
        "Continuous": pa.float64(),
        "Text": pa.string(),
        "Date": pa.string(),  # pa.date64(),
        "Time": pa.string(),  # pa.timestamp("s"),
        "Compound": pa.string(),
        "Categorical single": pa.string(),
        "Categorical multiple": pa.string(),
    }
    filename = "dictionary.parquet"

    def __init__(self, path: Path):
        """Constructor.

        Args:
            path: Directory to store the dictionary
        """
        self._fields = None
        self.path = path / self.filename

    def exists(self) -> bool:
        return self.path.exists()

    @property
    def fields(self):
        if self._fields is not None:
            return self._fields

        dictionary = self.load()
        dictionary["Type"] = dictionary.ValueType.map(self.field_types)
        self._fields = dictionary.set_index(dictionary.FieldID)
        return self._fields

    def filter(self, fields: List[str], search: str) -> pd.DataFrame:
        """Filter dictionary.

        Args:
            fields:
            search:

        Returns:
            None
        """
        dictionary = self.load()
        field_ids = set(map(self.get_field_id, fields))
        dictionary = dictionary[dictionary.FieldID.isin(field_ids)]
        if search:
            dictionary = dictionary[
                dictionary.Field.str.contains(search, case=False)
            ]
        return dictionary

    def load(self, path=None, download: bool = False) -> pd.DataFrame:
        if not self.exists() and download:
            self.download(path)

        dictionary = pd.read_parquet(self.path)
        dictionary.FieldID = dictionary.FieldID.astype(str)
        return dictionary

    def download(self, path=None) -> None:
        """Download dictionary from URL.

        Returns:
            None
        """
        if not path:
            path = settings.dictionary.url

        self.path.parent.mkdir(parents=True, exist_ok=True)
        with ProgressBar():
            print(f"loading dictionary from {path}")
            dictionary = dd.read_table(path)
            dictionary = dictionary.compute()
            dictionary.to_parquet(self.path)

    def get_field_id(self, field: str) -> str:
        """Get field ID.

        Args:
            field: Field name

        Returns:
            Field ID as str
        """
        return field.split("-")[0]

    def get_type(self, field) -> pa.DataType:
        """Get Arrow data type for a field.

        Args:
            field: Name of field

        Returns:
            Arrow field type
        """
        field = self.get_field_id(field)
        if field in self.fields.index:
            return self.fields.loc[field].Type

    def get_pandas_dtype(self, field) -> Any:
        """Get Pandas type for a field.

        Args:
            field: Name of field

        Returns:
            Pandas dtype
        """
        field_type = self.get_type(field)
        if not field_type:
            return None

        if field_type == pa.int64():
            return "Int64"  # use pandas nullable integer type

        return field_type.to_pandas_dtype()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['biobank', 'tests']

package_data = \
{'': ['*'], 'biobank': ['conf/*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'dask>=2022.2.0,<2023.0.0',
 'pandas>=1.4.1,<2.0.0',
 'pyarrow>=7.0.0,<8.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'requests>=2.27.1,<3.0.0',
 'typer==0.4.0']

extras_require = \
{'dev': ['tox>=3.20.1,<4.0.0',
         'virtualenv>=20.2.2,<21.0.0',
         'pip>=20.3.1,<21.0.0',
         'twine>=3.3.0,<4.0.0',
         'pre-commit>=2.12.0,<3.0.0',
         'toml>=0.10.2,<0.11.0',
         'bump2version>=1.0.1,<2.0.0'],
 'doc': ['mkdocs>=1.1.2,<2.0.0',
         'mkdocs-include-markdown-plugin>=1.0.0,<2.0.0',
         'mkdocs-material>=6.1.7,<7.0.0',
         'mkdocstrings>=0.15.2,<0.16.0',
         'mkdocs-autorefs>=0.2.1,<0.3.0'],
 'test': ['autoflake>=1.4,<2.0',
          'black>=21.5b2,<22.0',
          'flake8>=3.9.2,<4.0.0',
          'flake8-docstrings>=1.6.0,<2.0.0',
          'isort>=5.8.0,<6.0.0',
          'mypy>=0.900,<0.901',
          'pytest>=6.2.4,<7.0.0',
          'pytest-cov>=2.12.0,<3.0.0']}

entry_points = \
{'console_scripts': ['biobank = biobank.cli:commands']}

setup_kwargs = {
    'name': 'biobank-tools',
    'version': '0.1.0',
    'description': 'Biobank Tools.',
    'long_description': "# Biobank Tools\n\n\n[![pypi](https://img.shields.io/pypi/v/biobank-tools.svg)](https://pypi.org/project/biobank-tools/)\n[![python](https://img.shields.io/pypi/pyversions/biobank-tools.svg)](https://pypi.org/project/biobank-tools/)\n[![Build Status](https://github.com/altaf-ali/biobank-tools/actions/workflows/dev.yml/badge.svg)](https://github.com/altaf-ali/biobank-tools/actions/workflows/dev.yml)\n[![codecov](https://codecov.io/gh/altaf-ali/biobank-tools/branch/main/graphs/badge.svg)](https://codecov.io/github/altaf-ali/biobank-tools)\n\n* Documentation: <https://altaf-ali.github.io/biobank-tools>\n* GitHub: <https://github.com/altaf-ali/biobank-tools>\n* PyPI: <https://pypi.org/project/biobank-tools/>\n\n\n## Features\n\nThe Biobank Tools package provides simple, fast, and efficient access to UK\nBiobank data. Once you've downloaded and extracted the UK Biobank data to\ncomma or tab separated files, you can use Biobank Tools to convert the data\nto a format that's better suited for searching and filtering than plain text\nfiles. Internally, Biobank Tools convert the data to [Apache Parquet][]\nformat for optimized column-wise access.\n\n## Requirements\n\nBiobank Tools require Python 3.8 or above.\n\n## Credits\n\nThis package was created with [Cookiecutter][] and the [altaf-ali/cookiecutter-pypackage][] project template.\n\n[Apache Parquet]: https://parquet.apache.org/documentation/latest\n[Cookiecutter]: https://cookiecutter.readthedocs.io/en/latest/\n[altaf-ali/cookiecutter-pypackage]: https://altaf-ali.github.io/cookiecutter-pypackage\n",
    'author': 'Altaf Ali',
    'author_email': 'altaf@firecrest.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/altaf-ali/biobank-tools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Implementation of various file importers."""

from abc import ABC, abstractmethod
from pathlib import Path
from urllib.parse import urlsplit

import dask.dataframe as dd
import pyarrow as pa

from biobank.config import settings
from biobank.singleton import Singleton


class UnknownDataFormat(Exception):
    """Unknow dataset exception."""

    def __init__(self, suffix):
        self.suffix = suffix

    def __str__(self):
        """Converts the exception to a printable string."""
        return f"Unknown format '{self.suffix}'"


class ImportManager(metaclass=Singleton):
    """Import manager class."""

    def __init__(self):
        self.importers = dict()

    def register_importer(self, cls, suffix):
        self.importers[suffix] = cls

    @staticmethod
    def register(suffix):
        def registrar(cls):
            importer = ImportManager()
            importer.register_importer(cls, suffix)
            return cls

        return registrar

    def get_suffix(self, path):
        url = urlsplit(str(path))
        if url.scheme:
            path = url.path

        return Path(path).suffix

    def import_dataset(self, dictionary, path, **kwargs):
        print(f"importing {path}")
        suffix = self.get_suffix(path)
        try:
            importer = self.importers[suffix](dictionary)
            return importer(path, **kwargs)
        except KeyError:
            raise UnknownDataFormat(suffix)


class DataImporter(ABC):
    """Data importer abstract base class."""

    def __init__(self, dictionary):
        self.dictionary = dictionary

    @abstractmethod
    def __call__(self, path, **kwargs):
        raise NotImplementedError

    def get_schema(self, path, import_func):
        header = import_func(path)
        columns = dict.fromkeys(header.columns)
        schema = {
            col: self.dictionary.get_type(col) for col, _ in columns.items()
        }
        dtypes = {
            col: self.dictionary.get_pandas_dtype(col)
            for col, _ in columns.items()
        }
        schema[settings.fields.index] = pa.int64()
        return schema, dtypes

    def import_dataset(self, path, import_func):
        schema, dtypes = self.get_schema(path, import_func)
        dataset = import_func(path, dtype=dtypes)
        dataset = dataset.set_index(settings.fields.index)
        return dataset, schema


@ImportManager.register(".csv")
class CommaSeparated(DataImporter):
    """Comma separated file importer."""

    def __call__(self, path, **kwargs):
        return self.import_dataset(path, import_func=dd.read_csv)


@ImportManager.register(".tsv")
class TabSeparated(DataImporter):
    """Tab separated file importer."""

    def __call__(self, path, **kwargs):
        return self.import_dataset(path, import_func=dd.read_table)



"""Console script for biobank."""

from pathlib import Path
from typing import List

import typer

from biobank.dataset import Dataset


class AlertMarkers:
    """CLI alert markers."""

    error = typer.style("x ", fg=typer.colors.RED)
    success = typer.style("✓ ", fg=typer.colors.GREEN)
    warning = typer.style("! ", fg=typer.colors.YELLOW)


class Commands(typer.Typer):
    """CLI Command base class."""

    def __init__(self, **kwargs):
        """Constructor."""
        super().__init__(**kwargs)
        self.dataset = Dataset()

    def validate_dataset(self):
        if not self.dataset.path.exists():
            self.exit("Dataset does not exist. Please import a dataset first.")

    def exit(self, message, *args, **kwargs):
        self.alert_error(message, *args, **kwargs)
        raise typer.Exit()

    def echo(self, message, *args, **kwargs):
        typer.echo(message, *args, **kwargs)

    def alert(self, success, message, *args, **kwargs):
        if success:
            return self.alert_success(message, *args, **kwargs)
        else:
            return self.alert_error(message, *args, **kwargs)

    def alert_error(self, message, *args, **kwargs):
        self.echo(AlertMarkers.error + message, *args, **kwargs)

    def alert_success(self, message, *args, **kwargs):
        self.echo(AlertMarkers.success + message, *args, **kwargs)

    def alert_warning(self, message, *args, **kwargs):
        self.echo(AlertMarkers.warning + message, *args, **kwargs)


commands = Commands()


@commands.command(name="import")
def import_dataset(
    filename: Path, dictionary: Path = typer.Option(None), force: bool = False
) -> None:
    """Import biobank dataset."""
    if commands.dataset.path.exists():
        if force:
            commands.dataset.delete()
        else:
            commands.exit(
                "Dataset already imported. Use --force to overwite it."
            )
    commands.dataset.import_dataset(filename, dictionary)


@commands.command()
def exclude(files: List[Path], force: bool = False) -> None:
    """Exclude from biobank dataset."""
    commands.validate_dataset()
    commands.dataset.exclude(files)


@commands.command()
def select(
    fields: List[str] = typer.Argument(None),
    limit: int = None,
    output: Path = None,
) -> None:
    """Select fields from a dataset."""
    commands.validate_dataset()
    dataset = commands.dataset.select(fields, limit)
    if dataset.empty:
        commands.exit("No matching fields found.")

    if output:
        print(f"writing {output}")
        dataset.to_csv(output)
    else:
        print(dataset)


@commands.command()
def fields(
    fields: List[str] = typer.Argument(None),
    search: str = None,
    output: Path = None,
) -> None:
    """Show dataset fields."""
    commands.validate_dataset()
    dictionary = commands.dataset.filter_dictionary(fields, search)
    if dictionary.empty:
        commands.exit("No matching fields found.")

    if output:
        print(f"writing {output}")
        dictionary.to_csv(output, index=False)
    else:
        print(dictionary[["FieldID", "ValueType", "Units", "Field"]])


if __name__ == "__main__":
    commands()  # pragma: no cover



"""Biobank Dataset class."""

from pathlib import Path
from typing import Any, Tuple

import yaml
from pydantic import AnyUrl, BaseModel, BaseSettings
from pydantic.env_settings import SettingsSourceCallable


class FieldsConfig(BaseModel):
    """Fields config."""

    index: str


class DictionaryConfig(BaseModel):
    """Dictionary config."""

    url: AnyUrl


class Settings(BaseSettings):
    """Package settings."""

    debug: bool = False
    path: Path = Path(".biobank")
    fields: FieldsConfig
    dictionary: DictionaryConfig

    class Config:
        """Configuration loader."""

        config_dir = "conf"
        config_file = "config.yaml"

        @classmethod
        def file_settings(cls, settings: BaseSettings) -> dict[str, Any]:
            package_dir = Path(__file__).parent.absolute()
            config_file = package_dir / cls.config_dir / cls.config_file
            if config_file.is_file():
                with open(config_file) as f:
                    return yaml.safe_load(f)
            else:
                return settings.dict()

        @classmethod
        def customise_sources(
            cls,
            init_settings: SettingsSourceCallable,
            env_settings: SettingsSourceCallable,
            file_secret_settings: SettingsSourceCallable,
        ) -> Tuple[SettingsSourceCallable, ...]:
            return init_settings, cls.file_settings


settings = Settings()
